package ap.parameters;

import ap.util.{Debug, Seqs}

object Settings {
  
  private val AC = Debug.AC_PARAMETERS
  
}

object GlobalSettings {
  
  /**
   * An boolean option of the kind "+opt" or "-opt"
   */
  private object Opt {
    def unapply(arg : String) : Option[(String, Boolean)] =
      if (arg startsWith "+")
        Some(arg substring 1, true)
      else if (arg startsWith "-")
        Some((arg substring 1, false))
      else
        None
  }
  
  /**
   * An assignment of the kind "-var13=4" or "-xxx=hallo". The assigned value
   * is here only decoded to a string
   */  
  private object ValueOpt {
    def unapply(arg : String) : Option[(String, String)] =
      if ((arg startsWith "-") && (arg contains "=")) {
        val i = arg indexOf "="
        Some(arg.substring(1, i), arg substring (i+1))
      } else {
        None
      }
  }

  /**
   * A value that represents an integer, like in "-var13=4"
   */  
  private object IntVal {
    def unapply(arg : String) : Option[Int] =
      try {
       	Some(arg.toInt) 
      } catch {
        case _ : NumberFormatException => None
      }
  }
    
  def fromArguments(args : Array[String]) : (GlobalSettings, Seq[String]) = {
    var settings = DEFAULT
    val inputs = new scala.collection.mutable.ArrayBuffer[String]
    
    for (arg <- args)
      settings = arg match {
        case Opt("logo", value) =>
          Param.LOGO.set(settings, value)
        case Opt("printTree", value) =>
          Param.PRINT_TREE.set(settings, value)
        case Opt("assert", value) =>
          Param.ASSERTIONS.set(settings, value)
        case Opt("simplifyConstraints", value) =>
          Param.SIMPLIFY_CONSTRAINTS.set(settings, value)
        case Opt("traceConstraintSimplifier", value) =>
          Param.TRACE_CONSTRAINT_SIMPLIFIER.set(settings, value)
        case ValueOpt("timeout", IntVal(value)) =>
          Param.TIMEOUT.set(settings, value)
        case Opt("posUnitResolution", value) =>
          Param.POS_UNIT_RESOLUTION.set(settings, value)
        case Opt(_, _) =>
          throw new UnknownArgumentException(arg)
        case _ => { inputs += arg; settings }
      }
    
    (settings, inputs)
  }
  
  class UnknownArgumentException(arg : String)
        extends Exception("Unknown argument: " + arg)

  val allParams =
    List(Param.LOGO, Param.ASSERTIONS, Param.PRINT_TREE,
         Param.SIMPLIFY_CONSTRAINTS, Param.TRACE_CONSTRAINT_SIMPLIFIER,
         Param.TIMEOUT, Param.POS_UNIT_RESOLUTION)

  val DEFAULT =
    new GlobalSettings (scala.collection.immutable.HashMap[Param, Any]())

}

object GoalSettings {

  val allParams = List(Param.POS_UNIT_RESOLUTION)

  val DEFAULT =
    new GoalSettings (scala.collection.immutable.HashMap[Param, Any]())
  
}

abstract class Settings[STT <: Settings[STT]]
                       (protected val paramMap : Map[Param, Any]) {

  // all parameters that are allowed for this kind of setting
  protected val allParams : List[Param]
  private lazy val allValues = for (p <- allParams) yield this(p)  
  
  protected def setParams(paramMap : Map[Param, Any]) : STT
  
  protected[parameters] def apply(p : Param) : Any = {
    //////////////////////////////////////////////////////////////////////////
    Debug.assertPre(Settings.AC, allParams contains p)
    //////////////////////////////////////////////////////////////////////////
    (paramMap get p) match {
      case Some(value : Any) => {
        //////////////////////////////////////////////////////////////////////////
        Debug.assertPost(Settings.AC, value.isInstanceOf[p.Value])
        //////////////////////////////////////////////////////////////////////////
        value
      }
      case None => p.defau
    }
  }
  
  protected[parameters] def +(paramPair : (Param, Any)) : STT = {
    val (p, value) = paramPair
    //////////////////////////////////////////////////////////////////////////
    Debug.assertPre(Settings.AC, (allParams contains p) &&
                                 value.isInstanceOf[p.Value])
    ////////////////////////////////////////////////////////////////////////////
    setParams(paramMap + paramPair)
  }
  
  override def equals(that : Any) : Boolean = that match {
    case that : Settings[STT] => this.allValues sameElements that.allValues
    case _ => false
  }

  private lazy val hashCodeVal = Seqs.computeHashCode(this.allValues, 18732, 11)

  override def hashCode = hashCodeVal
}

class GlobalSettings(_paramMap : Map[Param, Any])
      extends Settings[GlobalSettings](_paramMap) {
  protected val allParams = GlobalSettings.allParams
  
  protected def setParams(paramMap : Map[Param, Any]) =
    new GlobalSettings(paramMap)
    
  def toGoalSettings : GoalSettings = {
    var res = GoalSettings.DEFAULT
    
    for ((p, v) <- paramMap) {
      if (GoalSettings.allParams contains p)
        res = res + (p, v)
    }
    
    res
  }
}

class GoalSettings(_paramMap : Map[Param, Any])
      extends Settings[GoalSettings](_paramMap) {
  protected val allParams = GoalSettings.allParams
  
  protected def setParams(paramMap : Map[Param, Any]) =
    new GoalSettings(paramMap)
}