package ap.proof.goal;

import ap.terfor.{Formula, TermOrder}
import ap.terfor.conjunctions.{Conjunction, Quantifier, ClauseMatcher}
import ap.terfor.preds.{Atom, PredConj, Predicate}
import ap.util.{Debug, Seqs, Logic}

import scala.collection.mutable.ArrayBuffer

object AddFactsTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  def isCoveredFormula(f : Conjunction) : Boolean =
    !f.isFalse &&
    (f.isTrue || f.isLiteral ||
     (f.isNegatedConjunction && f.negatedConjs(0).negatedConjs.isEmpty)) && 
    f.quans.isEmpty

}

class AddFactsTask(_formula : Conjunction, _age : Int)
      extends FormulaTask(_formula, _age) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(AddFactsTask.AC, !formula.isFalse)
  //////////////////////////////////////////////////////////////////////////////

  val priority : Int = age +
                       (if (formula.predicates.isEmpty)
                          -20000
                        else
                          -500)
  /**
   * Perform the actual task (whatever needs to be done with <code>formula</code>)
   */
  protected def doTask(goal : Goal, ptf : ProofTreeFactory) : ProofTree =
    if (formula.isTrue) {
      // then the goal can be closed immediately
      ptf.updateGoal(Conjunction.FALSE, goal)
    } else if (formula.isLiteral) {
      addFacts(Conjunction.negate(formula, goal.order), goal, ptf)
    } else {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(AddFactsTask.AC, formula.isNegatedConjunction)
      //////////////////////////////////////////////////////////////////////////

      val disj = formula.negatedConjs(0)
      //////////////////////////////////////////////////////////////////////////
      // possibly existing quantifiers should have been pulled out and
      // should not occur at this point
      Debug.assertInt(AddFactsTask.AC,
                      disj.quans.isEmpty && disj.negatedConjs.isEmpty)
      //////////////////////////////////////////////////////////////////////////
            
      // the literals of the negated conjunction can be added as facts to
      // the goal
      addFacts(disj, goal, ptf)
    }

  private def addFacts(newFacts : Conjunction,
                       goal : Goal,
                       ptf : ProofTreeFactory) : ProofTree = {
    val updatedFacts = Conjunction.conj(Array(goal.facts, newFacts), goal.order)
    val newTasks =
      unifyAtoms(goal.facts.predConj, newFacts.predConj, goal) ++
      matchClauses(newFacts.predConj, goal)
    ptf.updateGoal(updatedFacts, newTasks, goal)
  }

  private def matchClauses(newFacts : PredConj, goal : Goal) : Seq[PrioritisedTask] = {
    val matcher = new ClauseMatcher(goal.facts.predConj, newFacts,
                                    goal.reduceWithFacts, goal.order)

    for (clause <- goal.clauses) {
      if (NegLitClauseTask isCoveredFormula clause)
        matcher matchClause clause
    }

    Seqs.toRandomAccess(
      for (matchedClause <- matcher.resolvedClauses.elements;
           task <- goal.formulaTasks(matchedClause).elements) yield task)
  }
  
  private def unifyAtoms(oldFacts : PredConj,
                         newFacts : PredConj,
                         goal : Goal) : Seq[PrioritisedTask] =
    unifyAtoms(oldFacts.positiveLits, newFacts.negativeLits, goal) ++
    unifyAtoms(oldFacts.negativeLits, newFacts.positiveLits, goal) ++
    unifyAtoms(newFacts.negativeLits, newFacts.positiveLits, goal)
  
  private def unifyAtoms(atoms1 : RandomAccessSeq[Atom],
                         atoms2 : RandomAccessSeq[Atom],
                         goal : Goal) : Seq[PrioritisedTask] =
    if (atoms1.size > atoms2.size) {
      unifyAtoms(atoms2, atoms1, goal)
    } else {
      var currentPred : Predicate = null
      var currentOtherAtoms : Seq[Atom] = null
      for (atom <- atoms1;
           task <- {
             if (atom.pred != currentPred) {
               currentPred = atom.pred
               currentOtherAtoms = findAtomsWithSamePred(atom, atoms2, goal.order)
             }
             unifyAtoms(atom, currentOtherAtoms, goal)
           }) yield task
    }
    
  private def unifyAtoms(atom : Atom, otherAtoms : Seq[Atom],
                         goal : Goal) : Seq[PrioritisedTask] = {
    for (a2 <- otherAtoms;
         t <- {
                val eqs = atom.unify(a2, goal.order)
                val conj = Conjunction.conj(eqs, goal.order)
                goal.formulaTasks(goal reduceWithFacts conj)
              })
    yield t
  }
  
  /**
   * Find all atoms in the sequence <code>otherAtoms</code> that have the same
   * predicate as <code>atom</code>
   */
  private def findAtomsWithSamePred
                         (atom : Atom, otherAtoms : RandomAccessSeq[Atom],
                          order : TermOrder) : RandomAccessSeq[Atom] = {
    ////////////////////////////////////////////////////////////////////////////
    def post(res : RandomAccessSeq[Atom]) = {
      Debug.assertPost(AddFactsTask.AC,
                       Logic.forall(for (a <- res.elements)
                                    yield (a.pred == atom.pred)) &&
                       (for (a <- otherAtoms; if (a.pred == atom.pred))
                        yield a).size == res.size)
      res
    }
    ////////////////////////////////////////////////////////////////////////////
     
    // we assume that the sequence of atoms is sorted
    implicit def orderAtom(thisA : Atom) =
      new Ordered[Atom] {
        def compare(thatA : Atom) : Int =
          order.compare(thatA, thisA)
      }

    val pred = atom.pred

    def findAllAtoms(i : Int) = {
      var lowerBound : Int = i
      var upperBound : Int = i+1
      while (lowerBound > 0 && otherAtoms(lowerBound-1).pred == pred)
        lowerBound = lowerBound - 1
      while (upperBound < otherAtoms.size && otherAtoms(upperBound).pred == pred)
        upperBound = upperBound + 1
      post(otherAtoms.slice(lowerBound, upperBound))
    }
    
    Seqs.binSearch(otherAtoms, 0, otherAtoms.size, atom) match {
    case Seqs.Found(i) =>
      findAllAtoms(i)
    case Seqs.NotFound(i) =>
      if (i > 0 && otherAtoms(i-1).pred == pred)
        findAllAtoms(i-1)
      else if (i < otherAtoms.size && otherAtoms(i).pred == pred)
        findAllAtoms(i)
      else
        // no other atoms with the same predicate exist
        post(otherAtoms.slice(0, 0))
    }
  }
  
  /**
   * Create a new <code>FormulaTask</code> by updating the value of
   * <code>formula</code>
   */
  protected def updateFormula(f : Conjunction) : FormulaTask =
    new AddFactsTask(f, age)

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  protected def isCoveredFormula(f : Conjunction) : Boolean =
    AddFactsTask.isCoveredFormula(f)

  protected val name : String = "AddFacts"

}
