package ap.proof.goal;

import ap.terfor.{Formula, ConstantTerm}
import ap.terfor.conjunctions.{Conjunction, NegatedConjunctions}
import ap.terfor.arithconj.ArithConj
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.inequalities.InEqConj
import ap.terfor.substitutions.VariableSubst
import ap.terfor.preds.PredConj
import ap.util.{Debug, PlainRange, Seqs}

object BetaFormulaTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK
  
  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  def isCoveredFormula(f : Conjunction) : Boolean =
    f.size > 1 && f.quans.isEmpty

}


class BetaFormulaTask(_formula : Conjunction, _age : Int)
      extends FormulaTask(_formula, _age) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(BetaFormulaTask.AC,
                   !formula.isTrue && !formula.isFalse &&
                   !formula.isLiteral && !formula.isNegatedConjunction)
  //////////////////////////////////////////////////////////////////////////////
      
  val priority : Int = -300 + age +
                       formula.arithConj.size * 50 +
                       formula.predConj.size * 10 +
                       formula.negatedConjs.size *50

  /**
   * Perform the actual task (whatever needs to be done with <code>formula</code>)
   */
  protected def doTask(goal : Goal, ptf : ProofTreeFactory) : ProofTree =
    if (!splittingNecessary(goal)) {
      // we just add the formula as a clause to the goal
      val newClauses = NegatedConjunctions(goal.clauses.elements ++
                                           Iterator.single(formula),
                                           goal.order)
      ptf.updateGoal(newClauses, goal)
    } else {
      if (formula.arithConj.isTrue && formula.predConj.isTrue) {
        // split to handle the complex conjuncts
        applyBeta(goal, ptf)
      } else {
        // cut over one of the literals
        applyCut(goal, ptf)
      }
    }

  /**
   * Determine whether this formula requires real splitting, or whether it can
   * be passed to the constraint unchanged
   */
  def splittingNecessary(goal : Goal) : Boolean =
    !formula.predicates.isEmpty ||
    !Conjunction.collectQuantifiers(formula).isEmpty ||
    !Seqs.disjoint(goal.eliminatedConstants, formula.constants)
      
  private def applyBeta(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, formula.negatedConjs.size > 1)
    ////////////////////////////////////////////////////////////////////////////
      
    val subtrees = for (negConj <- formula.negatedConjs)
                   yield ptf.updateGoal(goal.formulaTasks(negConj.negate), goal)
    ptf.and(subtrees, goal.order)
  }

  private def applyCut(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    val order = goal.order    
    val (cutLiteral, rem) = selectCutLiteral(formula, goal.eliminatedConstants)
    
    val firstTree = {
      // assume that the cut-literal is true
      val remTasks = goal.formulaTasks(rem)
      val allTasks =
        if (introduceLemma(cutLiteral, goal))
          // make a real cut
          remTasks ++ goal.formulaTasks(cutLiteral.negate)
        else
          remTasks
          
      ptf.updateGoal(allTasks, goal)
    }
    
    val secondTree = {
      // assume that the cut-literal is false
      ptf.updateGoal(goal.formulaTasks(cutLiteral), goal)
    }
    
    ptf.and(Array(secondTree, firstTree), order)    
  }
  
  private def introduceLemma(cutLiteral : Conjunction, goal : Goal) : Boolean =
    if (cutLiteral.predConj.isTrue)
      // the cut-literal is an arithmetic literal
      cutLiteral.arithConj.negativeEqs.isEmpty &&
      !Seqs.disjoint(goal.eliminatedConstants, cutLiteral.constants)
    else
      // the cut-literal is a predicate literal
      cutLiteral.constants subsetOf goal.eliminatedConstants
  
  /**
   * Given a conjunction of formulas, select one of the formulas
   * as literal to cut over. The result is the pair of the selected formula and
   * the remaining <code>Conjunction</code>
   */
  private def selectCutLiteral(conj : Conjunction,
                               eliminatedConstants : Set[ConstantTerm]) : (Conjunction, Conjunction) =
    if (conj.predConj.isTrue) {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(BetaFormulaTask.AC, !conj.arithConj.isTrue)
      //////////////////////////////////////////////////////////////////////////
      val (sel, rem) = selectCutLiteral(conj.arithConj, eliminatedConstants)
      (Conjunction.conj(sel, conj.order), conj.updateArithConj(rem, conj.order))      
    } else {
      val (sel, rem) = selectCutLiteral(conj.predConj)
      (Conjunction.conj(sel, conj.order), conj.updatePredConj(rem, conj.order))            
    }

  private def selectCutLiteral(ac : ArithConj,
                               eliminatedConstants : Set[ConstantTerm])
              : (ArithConj, ArithConj) = {
    def selectPositiveEq = {
      val (sel, rem) = selectCutLiteral(ac.positiveEqs)
      (ArithConj.conj(sel, ac.order), ac.updatePositiveEqs(rem, ac.order))      
    }
    def selectNegativeEq = {
      val (sel, rem) = selectCutLiteral(ac.negativeEqs)
      (ArithConj.conj(sel, ac.order), ac.updateNegativeEqs(rem, ac.order))      
    }
    def selectInEq = {
      val (sel, rem) = selectCutLiteral(ac.inEqs)
      (ArithConj.conj(sel, ac.order), ac.updateInEqs(rem, ac.order))      
    }
    
    // try to select a cut-literal that contains eliminated constants
    (ac.positiveEqs.isEmpty,
       Seqs.disjoint(eliminatedConstants, ac.positiveEqs.constants),
     ac.negativeEqs.isEmpty,
       Seqs.disjoint(eliminatedConstants, ac.negativeEqs.constants),
     ac.inEqs.isEmpty,
       Seqs.disjoint(eliminatedConstants, ac.inEqs.constants)) match {
      
      case (false, false, _, _, _, _) => selectPositiveEq
      case (_, _, _, _, false, false) => selectInEq
      case (_, _, false, false, _, _) => selectNegativeEq
      case (false, _, _, _, _, _)     => selectPositiveEq
      case (_, _, _, _, false, _)     => selectInEq
      case (_, _, false, _, _, _)     => selectNegativeEq
      
    }
  }

  private def selectCutLiteral(eqs : EquationConj) : (EquationConj, EquationConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !eqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    (EquationConj(eqs(0), eqs.order),
     eqs.updateEqsSubset(eqs.slice(1, eqs.size), eqs.order))
  }

  private def selectCutLiteral(eqs : NegEquationConj)
                                      : (NegEquationConj, NegEquationConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !eqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    (NegEquationConj(eqs(0), eqs.order),
     eqs.updateEqsSubset(eqs.slice(1, eqs.size), eqs.order))
  }
  
  private def selectCutLiteral(inEqs : InEqConj) : (InEqConj, InEqConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !inEqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    (InEqConj(inEqs(0), inEqs.order),
     InEqConj(inEqs.slice(1, inEqs.size), inEqs.order))
  }
  
  private def selectCutLiteral(conj : PredConj) : (PredConj, PredConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !conj.isTrue)
    ////////////////////////////////////////////////////////////////////////////
    val posLits = conj.positiveLits
    val negLits = conj.negativeLits
    if (!posLits.isEmpty)
      (PredConj(List(posLits(0)), List(), conj.order),
       PredConj(posLits.slice(1, posLits.size), negLits, conj.order))
    else
      (PredConj(List(), List(negLits(0)),conj.order),
       PredConj(List(), negLits.slice(1, negLits.size), conj.order))
  }

  /**
   * Create a new <code>FormulaTask</code> by updating the value of
   * <code>formula</code>
   */
  protected def updateFormula(f : Conjunction) : FormulaTask =
    new BetaFormulaTask(f, age)

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  protected def isCoveredFormula(f : Conjunction) : Boolean =
    BetaFormulaTask.isCoveredFormula(f)

  protected val name : String = "BetaFor"

}
