package ap.proof.goal;

import ap.basetypes.IdealInt
import ap.terfor.{Formula, ConstantTerm, VariableTerm}
import ap.terfor.conjunctions.{Conjunction, Quantifier, NegatedConjunctions}
import ap.terfor.arithconj.ArithConj
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.inequalities.InEqConj
import ap.util.{Debug, PlainRange, Seqs}

import scala.collection.mutable.ArrayBuffer

object DivisibilityTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK
  
}

class DivisibilityTask(_formula : Conjunction, _age : Int)
      extends FormulaTask(_formula, _age) {

  val priority : Int = -50 + age

  protected def doTask(goal : Goal, ptf : ProofTreeFactory) : ProofTree =
    if (!splittingNecessary(goal)) {
      // we just add the formula as a clause to the goal
      val newClauses = NegatedConjunctions(goal.clauses.elements ++
                                           Iterator.single(formula),
                                           goal.order)
      ptf.updateGoal(newClauses, goal)
    } else {
      splitDivisibility(goal, ptf)
    }
    
  /**
   * Transform a formula <code>EX (n*_0 + t = 0)</code> in the succedent into
   * a conjunction
   * <code>ALL (n*_0 + t + 1 != 0) /\ ALL (n*_0 + t + 2 != 0) /\ ...</code>
   * in the succedent
   */
  private def splitDivisibility(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    val lc = formula.arithConj.positiveEqs(0)
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertInt(DivisibilityTask.AC,
                    !lc.isEmpty && lc.leadingTerm == VariableTerm(0))
    ////////////////////////////////////////////////////////////////////////////
      
    val coeff : IdealInt = lc.leadingCoeff
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertInt(DivisibilityTask.AC, coeff.signum > 0)
    ////////////////////////////////////////////////////////////////////////////
      
    val order = goal.order
      
    val cases = new ArrayBuffer[Formula]
      
    var offset : IdealInt = IdealInt.ONE
    while (offset < coeff) {
      val lcWithOffset = LinearCombination.sum(Array((IdealInt.ONE, lc),
                                                     (offset, LinearCombination.ONE)),
                                               order)
      cases += NegEquationConj(lcWithOffset, order)
      offset = offset + IdealInt.ONE
    }
      
    val caseDistinction = Conjunction.quantify(Array(Quantifier.ALL),
                                               Conjunction.conj(cases, order),
                                               order)
      
    ptf.updateGoal(goal.formulaTasks(caseDistinction), goal)
  }
    
  /**
   * Determine whether this formula requires real splitting, or whether it can
   * be passed to the constraint unchanged
   */
  def splittingNecessary(goal : Goal) : Boolean =
    !Seqs.disjoint(goal.eliminatedConstants, formula.constants)
  
  /*
   * Create a new <code>FormulaTask</code> by updating the value of
   * <code>formula</code>
   */
  protected def updateFormula(f : Conjunction) : FormulaTask =
    new DivisibilityTask(f, age)

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  def isCoveredFormula(f : Conjunction) : Boolean = f.isDivisibility

  protected val name : String = "DivisibilityFor"

}
