package ap.util;

import scala.util.Sorting
import scala.collection.mutable.ArrayBuffer

object Seqs {

  private val AC = Debug.AC_SEQ_UTILS

  /**
   * Lexicographic comparison of two lists of things
   */
  def lexCompare[T <% Ordered[T]](it1 : Iterator[T], it2 : Iterator[T]) : Int = {
    while (true) {
      if (it1.hasNext) {
        if (!it2.hasNext)
          return 1      
      } else {
        if (it2.hasNext)
          return -1
        else
          return 0      
      }      
      
      val compRes = it1.next compare it2.next
      if (compRes != 0) return compRes
    }
    
    throw new Error // never reached
  }

  /**
   * Interpret the given integers as results of a <code>compare</code>
   * function (negative, zero, positive for less, equal, greater) and combine
   * them lexicographically
   */
  def lexCombineInts(int1 : Int, int2 : => Int) : Int = {
    if (int1 == 0) int2 else int1
  }
  
  def lexCombineInts(int1 : Int, _int2 : => Int, _int3 : => Int) : Int = {
    if (int1 == 0) {
      val int2 = _int2
      if (int2 == 0) {
        _int3
      } else {
        int2
      }      
    } else {
      int1
    }
  }
   
  def lexCombineInts(int1 : Int, _int2 : => Int,
                     _int3 : => Int, _int4 : => Int) : Int = {
    if (int1 == 0) {
      val int2 = _int2
      if (int2 == 0) {
        val int3 = _int3
        if (int3 == 0) {
          _int4
        } else {
          int3
        }      
      } else {
        int2
      }      
    } else {
      int1
    }
  }
   
  /**
   * Compute a polynomial hashcode for a sequence of things
   */
  def computeHashCode[A](a : Iterable[A], init : Int, multiplier : Int) =
    (init /: a)((hash, el) => hash * multiplier + el.hashCode)
    
  //////////////////////////////////////////////////////////////////////////////

  abstract class BS_Result
  case class Found(index : Int) extends BS_Result
  case class NotFound(nextBiggerElement : Int) extends BS_Result

  /**
   * Binary search for an element in a sorted random-access sequent. The result
   * is either <code>Found(i)</code>, where <code>i</code> is the index of some
   * occurrence of <code>wanted</code> in <code>seq</code>, or
   * <code>NotFound(i)</code>, where <code>i</code> is the index of the
   * next-bigger element in <code>seq</code>. Note, that elements are never
   * compared with <code>==<code>, only with <code>(a compare b) == 0</code> 
   */
  def binSearch[T <% Ordered[T]](seq : RandomAccessSeq[T],
                                 begin : Int, end : Int, wanted : T) : BS_Result = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC,
                    end >= begin &&
                    Logic.forall(begin, end - 1,
                                 (i:Int) => seq(i) <= seq(i+1)))
    def post(res:BS_Result) = {
      Debug.assertPost(AC,
                       res match {
                       case Found(i) => i >= begin && i < end &&
                         (seq(i) compare wanted) == 0
                       case NotFound(i) => i >= begin && i <= end &&
                         (i == begin || seq(i-1) < wanted) &&
                         (i == end || seq(i) > wanted)
                       })
      res
    }
    ////////////////////////////////////////////////////////////////////////////
    
    var lo : Int = begin
    var hi : Int = end
      
    while ( lo + 1 < hi ) {
      val mid = (lo + hi) / 2
      val c = (seq(mid) compare wanted)
      if (c < 0) {
        lo = mid + 1
      } else if (c > 0) {
        hi = mid
      } else {
	return post(Found(mid))
      }
    }
      
    if (lo == hi) {
      post(NotFound(hi))
    } else {
      val c = (seq(lo) compare wanted)
      if (c < 0) {
        post(NotFound(hi))
      } else if (c > 0) {
        post(NotFound(lo))
      } else {
        post(Found(lo))
      }
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
   * Remove all duplicates from a sorted sequence. It is assumed that duplicates
   * can only occur immediately following each other
   */
  def removeDuplicates[A](s : Seq[A]) : Seq[A] = {
    val it = s.elements
    if (it.hasNext) {
      val res = new ArrayBuffer[A]
      var prevEl = it.next
      res += prevEl
      
      for (el <- it) {
        if (el != prevEl) {
          prevEl = el
          res += el
        }
      }
      
      //////////////////////////////////////////////////////////////////////////
      Debug.assertPost(AC, Logic.forall(0, res.size,
                                        i => Logic.forall(i+1, res.size,
                                          j => res(i) != res(j))))
      //////////////////////////////////////////////////////////////////////////
      res
    } else {
      s
    }
  }

  //////////////////////////////////////////////////////////////////////////////

  abstract class FAS_RESULT[+A]
  case class FilteredSorted[A](res : Array[A]) extends FAS_RESULT[A]
  case object FoundBadElement extends FAS_RESULT[Nothing]

  /**
   * Filter a sequence of objects in order to detect the existence of certain
   * bad objects (<code>badEl</code>) and to remove certain unnecessary objects
   * (<code>skipEl</code>). If a bad element is found,
   * <code>FoundBadElement</code> is returned, otherwise a sorted array with the
   * elements that were kept is created and returned.
   */
  def filterAndSort[A](it : Iterator[A],
                       skipEl : A => Boolean, badEl : A => Boolean,
                       trafo : A => A,
                       comesBefore : (A, A) => Boolean)
                                                            : FAS_RESULT[A] = {
    val buf = new ArrayBuffer[A]
    for (el <- it) {
      if (badEl(el)) return FoundBadElement
      if (!skipEl(el)) buf += trafo(el)
    }

    val ar = buf.toArray
    Sorting.stableSort(ar, comesBefore)

    FilteredSorted(ar)
  }

  //////////////////////////////////////////////////////////////////////////////

  /**
   * Return the first <code>Some(x)</code> of the given sequence, or
   * <code>None</code> if none exists
   */
  def some[A](vals : Iterator[Option[A]]) : Option[A] = {
    while (vals.hasNext) {
      val n = vals.next
      if (n.isDefined) return n
    }
    None
  }
   
  def some[A](vals : Iterable[Option[A]]) : Option[A] =
    some(vals.elements)

  //////////////////////////////////////////////////////////////////////////////

  def disjoint[A](a : scala.collection.Set[A], b : Iterable[A]) : Boolean =
    Logic.forall(for (x <- b.elements) yield !(a contains x))
    
  //////////////////////////////////////////////////////////////////////////////

  def toRandomAccess[A](els : Iterator[A]) : RandomAccessSeq[A] = {
    val buf = new ArrayBuffer[A]
    buf ++= els
    buf.readOnly
  }
    
  def toRandomAccess[A](els : Iterable[A]) : RandomAccessSeq[A] =
    toRandomAccess(els.elements)

  //////////////////////////////////////////////////////////////////////////////

  /**
   * reduceLeft that also works for empty sequences
   */
  def reduceLeft[A](els : Iterator[A], f : (A, A) => A) : Option[A] =
    if (els.hasNext) {
      var res = els.next
      while (els.hasNext) res = f(res, els.next)
      Some(res)
    } else {
      None
    }

  /**
   * reduceLeft that also works for empty sequences
   */
  def reduceLeft[A](els : Iterable[A], f : (A, A) => A) : Option[A] =
    reduceLeft(els.elements, f)
    
  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Split a sequence of things into two sequences, one with all the elements
   * for which certain predicate holds, and one with the elements for which the
   * predicate does not hold
   */
  def split[A](els : Iterator[A], firstKind : (A) => Boolean)
                                  : (RandomAccessSeq[A], RandomAccessSeq[A]) = {
    val res1 = new ArrayBuffer[A]
    val res2 = new ArrayBuffer[A]
    while (els.hasNext) {
      val n = els.next
      if (firstKind(n))
        res1 += n
      else
        res2 += n
    }
    (res1, res2)
  }

  def split[A](els : Iterable[A], firstKind : (A) => Boolean)
                                : (RandomAccessSeq[A], RandomAccessSeq[A]) =
    split(els.elements, firstKind)

  //////////////////////////////////////////////////////////////////////////////

  /**
   * Determine whether <code>a</code> occurs in <code>b</code> as a sub-sequence
   */
  def subSeq[A](a : Iterator[A], b : Iterator[A]) : Boolean = {
    while (a.hasNext) {
      val const = a.next
      do {
        if (!b.hasNext) return false
      } while (const != b.next)
    }
    true    
  }

  /**
   * Determine whether <code>a</code> occurs in <code>b</code> as a sub-sequence
   */
  def subSeq[A](a : Iterator[A], aFilter : scala.collection.Set[A],
                b : Iterator[A]) : Boolean = {
    while (a.hasNext) {
      val const = a.next
      if (aFilter contains const) {
        do {
          if (!b.hasNext) return false
        } while (const != b.next)
      }
    }
    true    
  }
    
}
